/*
       Copyright 1989, 1990, 1993, 1994 Digital Equipment Corporation
                        All Rights Reserved

Permission to use, copy, and modify this software and its documentation is
hereby granted only under the following terms and conditions.  Both the
above copyright notice and this permission notice must appear in all copies
of the software, derivative works or modified versions, and any portions
thereof, and both notices must appear in supporting documentation.

Users of this software agree to the terms and conditions set forth herein,
and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
free right and license under any changes, enhancements or extensions 
made to the core functions of the software, including but not limited to 
those affording compatibility with other hardware or software 
environments, but excluding applications which incorporate this software.
Users further agree to use their best efforts to return to Digital any
such changes, enhancements or extensions that they make and inform Digital
of noteworthy uses of this software.  Correspondence should be provided
to Digital at:

                      Director of Licensing
                      Western Research Laboratory
                      Digital Equipment Corporation
                      250 University Avenue
                      Palo Alto, California  94301

This software may be distributed (but not offered for sale or transferred
for compensation) to third parties, provided such third parties agree to
abide by the terms and conditions of this notice.

THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: printsubs.c,v 2.5 1995/08/19 09:55:23 vixie Exp $";
#endif

/*
 * printsubs.c
 *
 * Debugging printout functions
 *
 * Modification history:
 *
 * 19 December 1988	Jeffrey Mogul/DECWRL
 *	Created.
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <net/gw_screen.h>

#include <signal.h>
#include <stdio.h>
#include <errno.h>

#include "screentab.h"
#include "funcproto.h"

static void PrintObjectSpec(struct ObjectSpec *osp);
static void PrintAddrSpec(struct AddrSpec *asp, int flags);
static void PrintPortSpec(struct PortSpec *psp, int flags);
static void PortValPrint(struct PortValue pval);
static void PrintTCPFlags(char *msg, unsigned int flgs);

static struct icmp dummy;
int MinIcmpLen = sizeof(dummy.icmp_type) + sizeof(dummy.icmp_code) +
		 sizeof(dummy.icmp_cksum) + sizeof(dummy.icmp_hun);

void
PrintNetmaskData(nmp)
register struct NetmaskData *nmp;
{
	printf("NetmaskData: mask for %s", inet_ntoa(nmp->network));
	printf(" is %s\n", inet_ntoa(nmp->mask));
}

void
PrintActionSpec(ap)
register struct ActionSpec *ap;
{
	printf("from [");
	PrintObjectSpec(&(ap->from));
	printf("] to [");
	PrintObjectSpec(&(ap->to));
	printf("]");
	if (ap->to.tcpyesflags)
		PrintTCPFlags(" flags ", ap->to.tcpyesflags);
	if (ap->to.tcpnoflags)
		PrintTCPFlags(" flags-not ", ap->to.tcpnoflags);
	if (ap->action & ASACTION_ACCEPT)
		printf(" accept");
	else 
		printf(" reject");
	if (ap->action & ASACTION_NOTIFY)
		printf(" notify");
	if (ap->action & ASACTION_LOG)
		printf(" log");
	printf("\n");
}

static void
PrintObjectSpec(osp)
register struct ObjectSpec *osp;
{
	PrintAddrSpec(&(osp->aspec), osp->flags);
	printf("/");
	PrintPortSpec(&(osp->pspec), osp->flags);
}

static void
PrintAddrSpec(asp, flags)
register struct AddrSpec *asp;
int flags;
{
	if (flags & OSF_NOTADDR)
	    printf("not ");
	switch (asp->addrtype) {
	case ASAT_ANY:
		printf("any host");
		break;
	case ASAT_NET:
		printf("net %s", inet_ntoa(asp->aval.network));
		break;
	case ASAT_SUBNET:
		printf("subnet %s", inet_ntoa(asp->aval.subnet));
		break;
	case ASAT_HOST:
		printf("host %s", inet_ntoa(asp->aval.host));
		break;
	case ASAT_ADDR:
		printf("addr %s mask 0x%08lx",
		       inet_ntoa(asp->aval.addr),
		       ntohl(asp->mask.s_addr));
		break;
	default:
		printf("bad addrtype");
		break;
	}
}

static void
PrintPortSpec(psp, flags)
register struct PortSpec *psp;
int flags;
{
	char *ICMPNumberToName();
	char *icmpname;

	if (flags & OSF_NOTPROTO)
	    printf("not ");
	switch (psp->proto) {
	case IPPROTO_UDP:
		printf("UDP ");
		if (flags & OSF_NOTPORT)
		    printf("not ");
		PortValPrint(psp->pval.ports);
		break;
	case IPPROTO_TCP:
		printf("TCP ");
		if (flags & OSF_NOTPORT)
		    printf("not ");
		PortValPrint(psp->pval.ports);
		break;
	case IPPROTO_ICMP:
		if (flags & OSF_NOTPORT)
		    printf("not ");
		if (psp->pval.code == ICMPV_ANY) {
		    printf("any ICMP");
		    break;
		}
		if (psp->pval.code == ICMPV_INFOTYPE) {
		    printf("any `informational' ICMP");
		    break;
		}
		icmpname = ICMPNumberToName(psp->pval.code);
		if (icmpname)
		    printf("ICMP %s", icmpname);
		else
		    printf("ICMP code %d", psp->pval.code);
		break;
	case 0:
		printf("any proto");
		PortValPrint(psp->pval.ports);
		break;
	default:
		printf("proto %d", psp->proto);
		break;
	}
}

static void
PortValPrint(pval)
struct PortValue pval;
{
	printf(" port");
	if (pval.first == IPPORT_FIRST && pval.last == IPPORT_LAST)
	    printf(" any");
	else if (pval.first == pval.last)
	    printf(" %u", pval.first);
	else
	    printf(" range %u through %u", pval.first, pval.last);
}

static void
PrintTCPFlags(msg, flgs)
char *msg;
unsigned int flgs;
{
	printf(msg);
	if (flgs & TH_ACK)
	    putchar('A');
	if (flgs & TH_FIN)
	    putchar('F');
	if (flgs & TH_PUSH)
	    putchar('P');
	if (flgs & TH_RST)
	    putchar('R');
	if (flgs & TH_SYN)
	    putchar('S');
	if (flgs & TH_URG)
	    putchar('U');
}


/*
 * Prints an IP header
 */
void
PrintIPHeader(tvp, bufp, len)
struct timeval *tvp;
char *bufp;
int len;
{
	struct ip *ipp;
	struct tcphdr *tcpp;
	struct udphdr *udpp;
	int remlen;
	char *ipdata;

	remlen = len - sizeof(*ipp);
	if (remlen < 0) {
	    printf("Runt, len %d\n", len);
	    return;
	}
	ipp = (struct ip *)bufp;
	ipdata = &(bufp[sizeof(*ipp)]);
	if (ipp->ip_hl != 5) {
	    printf("Options present, hl %d ", ipp->ip_hl);
	}
	if (ipp->ip_off) {
	    printf("Not first frag, off %d\n", ntohs(ipp->ip_off));
	    return;
	}
	
	/* looks like a keeper? */
	printf("(%lu.%06lu) ", tvp->tv_sec, tvp->tv_usec);
	printf("[%s] -> ", inet_ntoa(ipp->ip_src));
	printf("[%s]", inet_ntoa(ipp->ip_dst));

	switch (ipp->ip_p) {
	case IPPROTO_UDP:
	    if (remlen < sizeof(*udpp)) {
		printf("Runt UDP header (%d short)\n",
			sizeof(*udpp) - remlen);
		return;
	    }
	    udpp = (struct udphdr *)ipdata;
	    printf(" UDP (%d->%d)\n",
			ntohs(udpp->uh_sport), ntohs(udpp->uh_dport));
	    break;
	case IPPROTO_TCP:
	    if (remlen < sizeof(*tcpp)) {
		printf("Runt TCP header (%d short)\n",
			sizeof(*tcpp) - remlen);
		return;
	    }
	    tcpp = (struct tcphdr *)ipdata;
	    printf(" TCP (%d->%d)\n",
			ntohs(tcpp->th_sport), ntohs(tcpp->th_dport));
	    break;
	/* known but not handled protocols: */
	case IPPROTO_ICMP:
	    printf(" ICMP");
	    break;
	default:
	    printf(" Unknown protocol %d", ipp->ip_p);
	}
	printf("\n");
}

/*
 * Formats an IP header into a string
 */
void
FormatIPHeader(bufp, len, msgbuf)
char *bufp;
int len;
char *msgbuf;
{
	struct ip *ipp;
	struct tcphdr *tcpp;
	struct udphdr *udpp;
	int remlen;
	char *ipdata;
	static char srcstr[32];
	static char opstr[64];
	struct icmp *icmpp;
	char *icmpname;
	char *ICMPNumberToAbbrev();
	unsigned int tcpflags;

	remlen = len - sizeof(*ipp);
	if (remlen < 0) {
	    sprintf(msgbuf, "Runt, len %d", len);
	    return;
	}
	ipp = (struct ip *)bufp;
	ipdata = &(bufp[sizeof(*ipp)]);
	if (ipp->ip_hl != 5) {
	    sprintf(opstr, " Options present, hl %d", ipp->ip_hl);
	}
	else
	    opstr[0] = '\0';
	if (ipp->ip_off) {
	    sprintf(msgbuf, "Not first frag, off %d", ntohs(ipp->ip_off));
	    return;
	}
	
	/* looks like a keeper? */
	sprintf(srcstr, "%s", inet_ntoa(ipp->ip_src));

	switch (ipp->ip_p) {
	case IPPROTO_UDP:
	    if (remlen < sizeof(*udpp)) {
		sprintf(msgbuf, "Runt UDP header (%d short)",
			sizeof(*udpp) - remlen);
		return;
	    }
	    udpp = (struct udphdr *)ipdata;
	    sprintf(msgbuf,
		"UDP [%s]->[%s](%d->%d)%s", srcstr, inet_ntoa(ipp->ip_dst),
			ntohs(udpp->uh_sport), ntohs(udpp->uh_dport),
			opstr);
	    break;
	case IPPROTO_TCP:
	    if (remlen < sizeof(*tcpp)) {
		sprintf(msgbuf, "Runt TCP header (%d short)",
			sizeof(*tcpp) - remlen);
		return;
	    }
	    tcpp = (struct tcphdr *)ipdata;
	    tcpflags = tcpp->th_flags;
	    /* Only log "unusual" flags (FIN/SYN/RST/URG) or not ACK */
	    if ((tcpflags & (~(TH_ACK|TH_PUSH))) || ((tcpflags & TH_ACK) == 0))
		sprintf(msgbuf,
		    "TCP [%s]->[%s](%d->%d)%02x%s",
		    	    srcstr, inet_ntoa(ipp->ip_dst),
			    ntohs(tcpp->th_sport), ntohs(tcpp->th_dport),
			    tcpflags,
			    opstr);
	    else
		sprintf(msgbuf,
		    "TCP [%s]->[%s](%d->%d)%s", srcstr, inet_ntoa(ipp->ip_dst),
			    ntohs(tcpp->th_sport), ntohs(tcpp->th_dport),
			    opstr);
	    break;
	/* known but not handled protocols: */
	case IPPROTO_ICMP:
	    if (remlen < MinIcmpLen) {
		sprintf(msgbuf, "Runt TCP header (%d short)",
			MinIcmpLen - remlen);
		return;
	    }
	    icmpp = (struct icmp *)ipdata;
	    icmpname = ICMPNumberToAbbrev(icmpp->icmp_type);
	    if (icmpname)
	        sprintf(msgbuf, "ICMP %s [%s]->[%s]%s", icmpname,
			srcstr, inet_ntoa(ipp->ip_dst), opstr
			);
	    else
	        sprintf(msgbuf, "ICMP code-%d [%s]->[%s]%s", icmpp->icmp_type,
			srcstr, inet_ntoa(ipp->ip_dst), opstr
			);
	    break;
	default:
	    sprintf(msgbuf, "Unknown proto %d [%s]->[%s]%s", ipp->ip_p,
			srcstr, inet_ntoa(ipp->ip_dst), opstr
			);
	}
}

/*
 * Formats an "unpacked" IP header into a string
 */
void
FormatUnpackedIPHeader(uhp, msgbuf)
struct unpacked_hdrs *uhp;
char *msgbuf;
{
	static char srcstr[32];
	char *icmpname;
	char *ICMPNumberToAbbrev();

	sprintf(srcstr, "%s", inet_ntoa(uhp->src.addr));

	switch (uhp->proto) {
	case IPPROTO_UDP:
	    sprintf(msgbuf,
		"UDP [%s]->[%s](%d->%d)", srcstr, inet_ntoa(uhp->dst.addr),
			uhp->src.port, uhp->dst.port);
	    break;
	case IPPROTO_TCP:
	    sprintf(msgbuf,
		"TCP [%s]->[%s](%d->%d)", srcstr, inet_ntoa(uhp->dst.addr),
			uhp->src.port, uhp->dst.port);
	    break;
	/* known but not handled protocols: */
	case IPPROTO_ICMP:
	    icmpname = ICMPNumberToAbbrev(uhp->src.port);
	    if (icmpname)
	        sprintf(msgbuf, "ICMP %s [%s]->[%s]", icmpname,
			srcstr, inet_ntoa(uhp->dst.addr)
			);
	    else
	        sprintf(msgbuf, "ICMP code-%d [%s]->[%s]", uhp->src.port,
			srcstr, inet_ntoa(uhp->dst.addr)
			);
	    break;
	default:
	    sprintf(msgbuf, "Unknown proto %d [%s]->[%s]", uhp->proto,
			srcstr, inet_ntoa(uhp->dst.addr)
			);
	}
}
